class NotFoundError(Exception):
    """Raised when not found."""
    pass

class ElasticsearchError(Exception):
    """Raised elasticsearch Error."""
    pass

class DocPathsWithoutValuesAggregationError(Exception):
    """Raised when the user defines aggregations with include_values=False and include_doc_paths=True"""
    pass



import json
from typing import List

from elasticsearch_dsl import Q, Search

from texta_elastic.core import ElasticCore
from texta_elastic.exceptions import DocPathsWithoutValuesAggregationError

class ElasticAggregator:
    """
    Everything related to performing aggregations in Elasticsearch
    """
    EMPTY_QUERY = {"query": {"match_all": {}}}


    def __init__(self, field_data=[], indices=[], query=EMPTY_QUERY):
        """
        field_data: list of decoded fields
        indices: list of index names (strings)
        """
        self.core = ElasticCore()
        self.field_data = field_data
        self.indices = indices
        self.query = query


    def _get_indices_string(self):
        return ",".join(self.indices)


    def update_query(self, query):
        self.query = query


    def update_field_data(self, field_data):
        """
        Updates field data. Expects list of decoded fields.
        """
        self.field_data = field_data


    def _aggregate(self, agg_query):
        self.query["aggregations"] = agg_query
        response = self.core.es.search(index=self.indices, body=self.query)
        return response


    def facts(self, size=30, filter_by_fact_name=None, min_count=0, max_count=None, include_values=True, include_doc_path=False, exclude_zero_spans=False):
        """
        For retrieving entities (facts) from ES.
        """
        agg_query = {
            "facts": {
                "nested": {"path": "texta_facts"},
                "aggs": {
                    "facts": {
                        "terms": {"field": "texta_facts.fact", "size": 10000}  # Get as many types of facts as possible.
                    }
                }
            }
        }

        if not include_values and include_doc_path:
            raise DocPathsWithoutValuesAggregationError("Cannot include doc path without including values! Please set `include_values=True`.")


        # filter by name if fact name present
        if filter_by_fact_name:
            agg_query["facts"]["aggs"]["facts"]["terms"]["include"] = [filter_by_fact_name]

        if include_values:
            agg_query["facts"]["aggs"]["facts"]["aggs"] = {"fact_values": {"terms": {"field": "texta_facts.str_val", "size": size}}}
            if include_doc_path:
                agg_query["facts"]["aggs"]["facts"]["aggs"]["fact_values"]["aggs"] = {"doc_paths": {"terms": {"field": "texta_facts.doc_path", "size": size}}}

            if exclude_zero_spans:
                if "aggs" in agg_query["facts"]["aggs"]["facts"]["aggs"]["fact_values"]:
                    agg_query["facts"]["aggs"]["facts"]["aggs"]["fact_values"]["aggs"].update({"fact_spans": {"terms": {"field": "texta_facts.spans", "size": 1}}})
                else:
                    agg_query["facts"]["aggs"]["facts"]["aggs"]["fact_values"]["aggs"] = {"fact_spans": {"terms": {"field": "texta_facts.spans", "size": 1}}}

        if exclude_zero_spans and not include_values:
            agg_query["facts"]["aggs"]["facts"]["aggs"] = {"fact_spans": {"terms": {"field": "texta_facts.spans", "size": 1}}}

        response = self._aggregate(agg_query)
        aggregations = response["aggregations"]

        entities = {}

        zero_span = json.dumps([[0, 0]])

        if aggregations["facts"]["doc_count"] > 0:
            fact_names = aggregations["facts"]["facts"]["buckets"]
            for fact_type in fact_names:
                fact_name = fact_type["key"]
                entities[fact_name] = []
                if "fact_values" in fact_type:
                    for fact_value in fact_type["fact_values"]["buckets"]:

                        fact_value_key = fact_value["key"]
                        fact_value_count = fact_value["doc_count"]

                        if exclude_zero_spans:
                            fact_span = fact_value["fact_spans"]["buckets"][0]["key"]

                        if include_doc_path:
                            fact_value_doc_path = fact_value["doc_paths"]["buckets"][0]["key"]
                            entity_to_add = {"value": fact_value_key, "doc_path": fact_value_doc_path}
                        else:
                            entity_to_add = fact_value_key

                        if fact_value_key and fact_value_count > min_count:
                            if max_count and fact_value_count >= max_count:
                                continue

                            if exclude_zero_spans:
                                if fact_span != zero_span:
                                    entities[fact_name].append(entity_to_add)

                            else:
                                entities[fact_name].append(entity_to_add)
                elif "fact_spans" in fact_type:
                    fact_span = fact_type["fact_spans"]["buckets"][0]["key"]
                    if fact_span == zero_span:
                        del entities[fact_name]


        # filter by name if fact name present
        if filter_by_fact_name:
            if filter_by_fact_name in entities:
                entities = entities[filter_by_fact_name]
            else:
                entities = []

        if not include_values:
            entities = list(entities.keys())
        return entities


    def filter_aggregation_maker(self, agg_type: str, field: str, filter_query: dict = None, size=1000, return_size=15, stop_words: List = None, exclude=""):

        container = []

        s = Search(using=self.core.es, index=self._get_indices_string())
        if filter_query:
            filter_query = Q(filter_query)
            s.aggs.bucket("limits", "filter", filter=filter_query).bucket("placekeeper", agg_type, field=field, size=size, exclude=exclude)
            r = s.execute()
            for hit in r.aggs.limits.placekeeper:
                container.append(hit.to_dict())
        else:
            s.aggs.bucket("placekeeper", agg_type, field=field)
            r = s.execute()
            for hit in r.aggs.limits.placekeeper:
                container.append(hit.to_dict())

        if stop_words:
            container = [item for item in container if item["key"] not in set(stop_words)]

        return container[:return_size]


    def get_significant_words(self, document_ids: List[str], field: str, stop_words: List = None, exclude="") -> List[dict]:
        """
        Args:
            stop_words: Optional parameter to remove stop words from significant words.
            document_ids: List of document ids to limit the range of the significant words.
            field: Path name of the field we're comparing text from for significant words.
            exclude: regex string for which values to exclude.

        Returns: List of dicts with the aggregation results.

        """
        query = {'ids': {'values': document_ids}}
        sw = self.filter_aggregation_maker(agg_type="significant_text", field=field, filter_query=query, stop_words=stop_words, exclude=exclude)
        sw = [{"key": hit["key"], "count": hit["doc_count"]} for hit in sw]
        return sw


    def get_fact_values_distribution(self, fact_name: str, fact_name_size: int = 30, fact_value_size: int = 30):
        """
        Returns a dictionary with fact values (labels) as keys and labels' quantities as values.
        """
        self.query["size"] = 0

        agg_query = {'fact_values': {'nested': {'path': 'texta_facts'},
                                     'aggs': {'fact_bucket': {'terms': {'field': 'texta_facts.fact', 'size': fact_name_size},
                                                              'aggs': {'value_bucket': {'terms': {'field': 'texta_facts.str_val', 'size': fact_value_size}}}}}}}
        # agg_query["size"] = 30

        fact_dist_dict = {}
        response = self._aggregate(agg_query)

        for fact in response["aggregations"]["fact_values"]["fact_bucket"]["buckets"]:
            if fact["key"] == fact_name:
                for fact_value in fact["value_bucket"]["buckets"]:
                    fact_dist_dict[fact_value["key"]] = fact_value["doc_count"]
        return fact_dist_dict



import os

import environ


env = environ.Env()

ES_CONNECTION_PARAMETERS = {
    "use_ssl": env.bool("TEXTA_ES_USE_SSL", default=False),
    "verify_certs": env.bool("TEXTA_ES_VERIFY_CERTS", default=False),
    "ca_certs": env("TEXTA_ES_CA_CERT_PATH", default=None),
    "client_cert": env("TEXTA_ES_CLIENT_CERT_PATH", default=None),
    "client_key": env("TEXTA_ES_CLIENT_KEY_PATH", default=None),
    "timeout": env.int("TEXTA_ES_TIMEOUT", default=60),
    "sniff_on_start": env.bool("TEXTA_ES_SNIFF_ON_START", default=True),
    "sniff_on_connection_fail": env.bool("TEXTA_ES_SNIFF_ON_FAIL", default=True)
}

TEXTA_ANNOTATOR_KEY = "texta_annotator"

TEXTA_ES_URL = env("TEXTA_ES_URL", default="http://localhost:9200")
TEXTA_ES_PREFIX = env("TEXTA_ES_PREFIX", default="")
TEXTA_ES_USERNAME = env("TEXTA_ES_USER", default="")
TEXTA_ES_PASSWORD = env("TEXTA_ES_PASSWORD", default="")
TEXTA_EVALUATOR_MEMORY_BUFFER_GB = env("TEXTA_EVALUATOR_MEMORY_BUFFER_GB", default="")

ERROR_LOGGER = "error_logger"

TEXTA_TAGS_KEY = "texta_facts"

TEST_INDEX = "texta_test_index"

TEST_FIELD = "comment_content_lemmas"

TEST_FIELD_CHOICE = [TEST_FIELD]

TEST_FACT_NAME = "TEEMA"

BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

# data dir for files generated by TK
DATA_DIR = os.getenv("TEXTA_DATA_DIR", default=os.path.join(BASE_DIR, "data"))

# Path to the directory containing test files
TEST_DATA_DIR = os.path.join(DATA_DIR, "test")



import datetime
import json
import logging
import uuid
from typing import List

import elasticsearch
import elasticsearch_dsl
from elasticsearch.helpers import bulk
from elasticsearch_dsl import Q, Search

from texta_elastic.decorators import elastic_connection
from texta_elastic.core import ElasticCore
from texta_elastic.searcher import EMPTY_QUERY
from texta_elastic.settings import ERROR_LOGGER, TEXTA_ANNOTATOR_KEY


class ESDocObject:
    """
    An object connected to ES document. Retrieves the document from ES on init.
    """


    def __init__(self, document_id=None, index: str = None, document=None):
        self.core = ElasticCore()
        self.document_id = document_id
        self.index = index
        self.document = self.get() if document_id else document


    @elastic_connection
    def get(self):
        """
        Retrieve document by ID.
        """
        document = self.core.es.get(self.index, self.document_id)
        return {
            "_index": document["_index"],
            "_type": document["_type"],
            "_id": document["_id"],
            "_source": document["_source"]
        }


    @staticmethod
    @elastic_connection
    def random_document(indices, query=EMPTY_QUERY):
        ec = ElasticCore()
        s = elasticsearch_dsl.Search(using=ec.es, index=indices)
        query = query.get("query", None) or query
        existing_query = Q(query)  # TODO Depending on whether you pull annotated or normal this becomes messy, fix later.
        random_query = Q("function_score", random_score={})
        s = s.query("bool", must=[existing_query, random_query])
        s = s.source(exclude=["*"])
        s = s.extra(size=1)
        hits = s.execute()
        for hit in hits:
            return ESDocObject(document_id=hit.meta.id, index=hit.meta.index)
        return None


    def apply_mlp(self, mlp, analyzers: List[str], field_data: List[str]):
        """
        Applies MLP to the selected fields and combines the results.
        """
        document_source = self.document["_source"]
        mlp_processed = mlp.process_docs([document_source], analyzers=analyzers, doc_paths=field_data)[0]
        self.document["_source"] = {**document_source, **mlp_processed}
        return True


    def add_field(self, field_name, field_content):
        """
        Adds field to document source.
        """
        self.document["_source"][field_name] = field_content
        return True


    @staticmethod
    def generate_fact_template(source="annotator"):
        """
        Pre-generates some fields to be added as a fact in the context of the annotator.
        :return:
        """
        return {"id": str(uuid.uuid4()), "source": source}


    def add_fact(self, fact_name, fact_value, doc_path, spans=json.dumps([[0, 0]]), sent_index=0, author="", source="annotator"):
        template_fact = self.generate_fact_template(source=source)
        fact = {**template_fact, "str_val": fact_value, "fact": fact_name, "doc_path": doc_path, "spans": spans, "sent_index": sent_index, "author": author}
        existing_facts = self.document["_source"].get("texta_facts", [])
        existing_facts.append(fact)
        existing_facts = ElasticDocument.remove_duplicate_facts(existing_facts)
        self.document["_source"]["texta_facts"] = existing_facts
        return fact

    def filter_facts(self, fact_name, doc_path):
        new_facts = []
        existing_facts = self.document["_source"].get("texta_facts", [])

        for fact in existing_facts:
            if fact["fact"] != fact_name or fact["doc_path"] != doc_path:
                new_facts.append(fact)

        self.document["_source"]["texta_facts"] = new_facts

        return new_facts



    # TODO These three can be unified into a more general function.
    def add_skipped(self, annotator_model, user):
        source = self.document["_source"]
        annotation_dicts = source.get(TEXTA_ANNOTATOR_KEY, [])
        job_dicts = [annotation_dict for annotation_dict in annotation_dicts if annotation_dict.get("job_id", None)]
        job_dict = job_dicts[0] if len(job_dicts) > 0 else {}
        job_dict["job_id"] = annotator_model.pk
        job_dict["user"] = user.username
        job_dict["skipped_timestamp_utc"] = datetime.datetime.utcnow()
        if not job_dicts:
            annotation_dicts.append(job_dict)
        self.document["_source"][TEXTA_ANNOTATOR_KEY] = annotation_dicts


    def add_validated(self):
        source = self.document["_source"]
        annotation_dict = source.get(TEXTA_ANNOTATOR_KEY, {})
        annotation_dict["validated_timestamp_utc"] = datetime.datetime.utcnow()
        self.document["_source"][TEXTA_ANNOTATOR_KEY] = annotation_dict


    def add_annotated(self, annotator_model, user):
        source = self.document["_source"]
        annotation_dicts = source.get(TEXTA_ANNOTATOR_KEY, [])
        job_dict = {}
        job_dict["job_id"] = annotator_model.pk
        job_dict["user"] = user.username
        job_dict["processed_timestamp_utc"] = datetime.datetime.utcnow()
        annotation_dicts.append(job_dict)
        self.document["_source"][TEXTA_ANNOTATOR_KEY] = annotation_dicts


    @elastic_connection
    def update(self, retry_on_conflict=3):
        """
        Updates document in ES by ID.
        """
        return self.core.es.update(
            index=self.document["_index"],
            doc_type=self.document["_type"],
            id=self.document["_id"] or self.document_id,
            body={"doc": self.document["_source"]},
            refresh="wait_for",
            retry_on_conflict=retry_on_conflict
        )


    @elastic_connection
    def delete(self):
        """
        Removes given document from ES.
        """
        return self.core.es.delete(index=self.index, id=self.document_id)


# This class should be modified for bulk operations etc
class ElasticDocument:
    """
    Everything related to managing documents in Elasticsearch
    """


    def __init__(self, index):
        self.core = ElasticCore()
        self.index = index


    @staticmethod
    def remove_duplicate_facts(facts: List[dict]):
        if facts:
            set_of_jsons = {json.dumps(fact, sort_keys=True, ensure_ascii=False) for fact in facts}
            without_duplicates = [json.loads(unique_fact) for unique_fact in set_of_jsons]
            return without_duplicates
        else:
            return []


    def __does_fact_exist(self, fact: dict, existing_facts: List[dict]):
        existing = {json.dumps(d, sort_keys=True, ensure_ascii=True) for d in existing_facts}
        checking = json.dumps(fact, sort_keys=True, ensure_ascii=True)
        if checking in existing:
            return True
        else:
            return False


    def _fact_addition_generator(self, documents, fact):
        for document in documents:
            # If there is no texta_facts field in the index, add it.
            if "texta_facts" not in document["_source"]:
                self.core.add_texta_facts_mapping(document["_index"])
                facts = document["_source"].get("texta_facts", [fact])
                doc_type = document.get("_type", "_doc")
                yield {
                    "_op_type": "update",
                    "retry_on_conflict": 3,
                    "_index": document["_index"],
                    "_type": doc_type,
                    "_id": document["_id"],
                    "doc": {"texta_facts": facts}
                }

            else:
                # Avoid sending duplicates.
                if self.__does_fact_exist(fact, document["_source"]["texta_facts"]):
                    pass
                else:
                    facts = document["_source"]["texta_facts"]
                    facts.append(fact)
                    doc_type = document.get("_type", "_doc")
                    yield {
                        "_op_type": "update",
                        "retry_on_conflict": 3,
                        "_index": document["_index"],
                        "_type": doc_type,
                        "_id": document["_id"],
                        "doc": {"texta_facts": facts}
                    }


    @elastic_connection
    def add_fact_to_documents(self, fact: dict, doc_ids: List):
        # Fetch the documents with the bulk function to get the facts,
        # and to validate that those ids also exist.
        documents = self.get_bulk(doc_ids=doc_ids, fields=["texta_facts"])
        generator = self._fact_addition_generator(documents, fact)
        self.bulk_update(generator)
        return True


    @elastic_connection
    def get(self, doc_id, fields: List = None):
        """
        Retrieve document by ID.
        """
        s = Search(using=self.core.es, index=self.index)
        s = s.query("ids", values=[doc_id])
        s = s.source(fields)
        s = s[:1000]
        response = s.execute()
        if response:
            document = response[0]
            doc_type = getattr(document.meta, "doc_type", "_doc")
            return {"_index": document.meta.index, "_type": doc_type, "_id": document.meta.id, "_source": document.to_dict()}
        else:
            return None


    @elastic_connection
    def get_bulk(self, doc_ids: List[str], fields: List[str] = None, flatten: bool = False) -> List[dict]:
        """
        Retrieve full Elasticsearch documents by their ids that includes id, index,
        type and content information. For efficiency it's recommended to limit the returned
        fields as unneeded content consumes extra internet bandwidth.
        """
        s = Search(using=self.core.es, index=self.index)
        s = s.query("ids", values=doc_ids)
        s = s.source(fields)
        s = s[:10000]
        response = s.execute()
        if response:
            container = []
            for document in response:
                document = {
                    "_index": document.meta.index,
                    "_type": getattr(document.meta, "doc_type", "_doc"),
                    "_id": document.meta.id,
                    "_source": self.core.flatten(document.to_dict()) if flatten else document.to_dict()
                }
                container.append(document)
            return container
        else:
            return []


    @elastic_connection
    def update(self, index, doc_id, doc, doc_type="_doc", retry_on_conflict=3):
        """
        Updates document in ES by ID.
        """
        return self.core.es.update(index=index, doc_type=doc_type, id=doc_id, body={"doc": doc}, refresh="wait_for", retry_on_conflict=retry_on_conflict)


    @elastic_connection
    def bulk_update(self, actions, refresh="wait_for", chunk_size=100):
        """
        Intermediary function to commit bulk updates.
        This function doesn't have actions processing because it's easier to use
        when it's index unaware. Actions should be processed when needed.

        Setting refresh to "wait_for" makes Python wait until the documents are actually indexed
        to avoid version conflicts.

        Args:
            chunk_size: How many documents should be sent per batch.
            refresh: Which behaviour to use for updating the index contents on a shard level.
            actions: List of dictionaries or its generator containing raw Elasticsearch documents along with
            a "doc" and "op_type" field that contains the fields that need updating. For ex:
            {"_id": 1234, "_index": "reddit", "op_type": "update", "doc": {"texta_facts": []}}

        Returns: Elasticsearch response to the request.
        """
        actions = self.add_type_to_docs(actions)
        return bulk(client=self.core.es, actions=actions, refresh=refresh, request_timeout=30, chunk_size=chunk_size)


    @elastic_connection
    def add(self, doc):
        """
        Adds document to ES.
        """
        return self.core.es.index(index=self.index, body=doc, refresh='wait_for')


    @elastic_connection
    def bulk_add(self, docs, chunk_size=100, raise_on_error=True, stats_only=True):
        actions = [{"_index": self.index, "_source": doc, "_type": doc.get("_type", "_doc")} for doc in docs]
        return bulk(client=self.core.es, actions=actions, chunk_size=chunk_size, stats_only=stats_only, raise_on_error=raise_on_error)


    def add_type_to_docs(self, actions):
        for action in actions:
            doc_type = action.get("_type", "_doc")
            action["_type"] = doc_type
            yield action


    @elastic_connection
    def bulk_add_generator(self, actions, chunk_size=100, raise_on_error=True, stats_only=True, refresh="wait_for"):
        actions = self.add_type_to_docs(actions)
        try:
            return bulk(client=self.core.es, actions=actions, chunk_size=chunk_size, stats_only=stats_only, raise_on_error=raise_on_error, refresh=refresh)
        except elasticsearch.helpers.errors.BulkIndexError as e:
            logging.getLogger(ERROR_LOGGER).exception(e.args[1][0]['index']['error']['reason'], exc_info=False)
            return None


    @elastic_connection
    def delete(self, doc_id):
        """
        Removes given document from ES.
        """
        return self.core.es.delete(index=self.index, id=doc_id)


    @elastic_connection
    def delete_by_query(self, query):
        """
        Removes given document from ES.
        """
        return self.core.es.delete_by_query(index=self.index, body=query)


    @elastic_connection
    def bulk_delete(self, document_ids: List[str], wait_for_completion=True):
        query = Search().query(Q("ids", values=document_ids)).to_dict()
        response = self.core.es.delete_by_query(index=self.index, body=query, wait_for_completion=wait_for_completion)
        return response


    @elastic_connection
    def count(self, indices=None) -> int:
        """
        Returns the document count for given indices.
        :indices: Either a coma separated string of indices or a list of index strings.
        """
        index = indices if indices else self.index
        return self.core.es.count(index=index)["count"]



import os
from setuptools import setup

def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "texta-elastic-tools",
    version = read("VERSION").strip(),
    author = "TEXTA",
    author_email = "info@texta.ee",
    description = ("TEXTA Elasticsearch tools"),
    license = "GPLv3",
    packages = ["texta_elastic"],
    data_files = ["VERSION", "requirements.txt", "README.md", "LICENSE"],
    long_description = read("README.md"),
    long_description_content_type="text/markdown",
    url="https://git.texta.ee/texta/texta-elastic-tools-python",
    install_requires = read("requirements.txt").split("\n"),
    include_package_data = True,
    extras_require = {
        'mlp': ['texta-mlp']
    }
)



from collections import defaultdict
from texta_elastic.mapping_generator import SchemaGenerator
from texta_elastic.core import ElasticCore


def get_selected_fields(indices, fields):
    # get all fields in given indices
    all_fields = ElasticCore().get_fields(indices)
    # filter out selected fields
    selected_fields = [field for field in all_fields if field["path"] in fields]
    return selected_fields


def add_nested(fields):
    for field in fields:
        if '.' in field['path']:
            field['type'] = 'nested'
    return fields


def reformat_for_schema_generator(fields, flatten_doc=False):
    if not flatten_doc:
        fields = add_nested(fields)
    formatted_fields = defaultdict(list)
    for field in fields:
        if field['path'] == 'texta_facts':
            formatted_fields['texta_facts'].append('texta_facts')
        elif field['path'] == 'texta_meta':
            formatted_fields['texta_meta'].append('texta_meta')
        else:
            formatted_fields[field['type']].append(field['path'])
    return dict(formatted_fields)


def update_field_types(indices, fields, field_type, flatten_doc=False):
    ''' if fieldtype, for field named fieldtype change its type'''

    # returns fields edited by serializer input
    my_fields = get_selected_fields(indices, fields)
    my_field_data = [field["path"] for field in my_fields]

    for item in field_type:
        if item['path'] in my_field_data:
            field_to_edit = item['path']
            new_type = item['field_type']

            for field in my_fields:
                if field['path'] == field_to_edit:
                    field['type'] = new_type
                    # TODO must work, also when only name is changed
                    if 'new_path_name' in item.keys():
                        new_path_name = item['new_path_name']
                        field['path'] = new_path_name
    updated_field_types = reformat_for_schema_generator(my_fields, flatten_doc)
    return updated_field_types


def update_mapping(schema_input, doc_type: str, add_facts_mapping, add_texta_meta_mapping):
    mod_schema = SchemaGenerator().generate_schema(schema_input, add_facts_mapping, add_texta_meta_mapping)
    return {'mappings': {"_doc": mod_schema}}
